@@ -1,8 +1,10 @@
+from __future__ import annotations
 import socketio
 import logging
 import time
 import datetime
 import math
+from inspect import signature
 from typing import overload, Union, TypedDict, Literal, Callable, List, Dict, Optional
 
 
@@ -40,10 +42,10 @@ EVENTS = Literal[
 
 
 class Device(TypedDict):
-    deviceId: str
-    isController: bool
-    deviceNr: int
-    socketId: str
+    device_id: str
+    is_controller: bool
+    device_nr: int
+    socket_id: str
 
 
 class DeviceJoinedMsg(TypedDict):
@@ -57,18 +59,18 @@ class DeviceLeftMsg(TypedDict):
 
 
 class TimeStampedMsg(TypedDict):
-    timeStamp: int
+    time_stamp: int
 
 
 class BaseMsg(TimeStampedMsg):
-    deviceId: str
-    deviceNr: int
+    device_id: str
+    device_nr: int
 
 
 class BaseSendMsg(TypedDict):
-    deviceId: Optional[str]
-    deviceNr: Optional[int]
-    timeStamp: Optional[int]
+    device_id: Optional[str]
+    device_nr: Optional[int]
+    time_stamp: Optional[int]
 
 
 class DataMsg(BaseMsg):
@@ -140,13 +142,13 @@ def flatten(list_of_lists: List[List]) -> List:
 
 def to_datetime(data: TimeStampedMsg) -> datetime.datetime:
     '''
-    extracts the datetime from a data package. if the field `timeStamp` is not present,
+    extracts the datetime from a data package. if the field `time_stamp` is not present,
     the current datetime will be returned
     '''
-    if 'timeStamp' not in data:
+    if 'time_stamp' not in data:
         return datetime.datetime.now()
-    ts = data['timeStamp']
-    # convert timestamp from ms to seconds
+    ts = data['time_stamp']
+    # convert time_stamp from ms to seconds
     if ts > 1000000000000:
         ts = ts / 1000.0
     return datetime.datetime.fromtimestamp(ts)
@@ -164,21 +166,21 @@ class Connector:
     joined_rooms: List[str]
 
     # callback functions
-    on_key: Callable[[KeyMsg], None] = None
-    on_pointer: Callable[[PointerMsg], None] = None
-    on_acceleration: Callable[[AccMsg], None] = None
-    on_gyro: Callable[[GyroMsg], None] = None
-    on_sensor: Callable[[Union[GyroMsg, AccMsg]], None] = None
+    on_key: Callable[[KeyMsg, Optional[Connector]], None] = None
+    on_pointer: Callable[[PointerMsg, Optional[Connector]], None] = None
+    on_acceleration: Callable[[AccMsg, Optional[Connector]], None] = None
+    on_gyro: Callable[[GyroMsg, Optional[Connector]], None] = None
+    on_sensor: Callable[[Union[GyroMsg, AccMsg], Optional[Connector]], None] = None
 
-    on_data: Callable[[DataMsg], None] = None
-    on_broadcast_data: Callable[[DataMsg], None] = None
-    on_all_data: Callable[[List[DataMsg]], None] = None
-    on_device: Callable[[Device], None] = None
-    on_client_device: Callable[[Union[Device, None]], None] = None
-    on_devices: Callable[[List[Device]], None] = None
-    on_error: Callable[[ErrorMsg], None] = None
-    on_room_joined: Callable[[DeviceJoinedMsg], None] = None
-    on_room_left: Callable[[DeviceLeftMsg], None] = None
+    on_data: Callable[[DataMsg, Optional[Connector]], None] = None
+    on_broadcast_data: Callable[[DataMsg, Optional[Connector]], None] = None
+    on_all_data: Callable[[List[DataMsg], Optional[Connector]], None] = None
+    on_device: Callable[[Device, Optional[Connector]], None] = None
+    on_client_device: Callable[[Union[Device, None], Optional[Connector]], None] = None
+    on_devices: Callable[[List[Device], Optional[Connector]], None] = None
+    on_error: Callable[[ErrorMsg, Optional[Connector]], None] = None
+    on_room_joined: Callable[[DeviceJoinedMsg, Optional[Connector]], None] = None
+    on_room_left: Callable[[DeviceLeftMsg, Optional[Connector]], None] = None
 
     @property
     def server_url(self):
@@ -205,7 +207,7 @@ class Connector:
 
     @property
     def client_device(self):
-        return next((device for device in self.devices if device['isController'] and device['deviceId'] == self.device_id), None)
+        return next((device for device in self.devices if device['is_controller'] and device['device_id'] == self.device_id), None)
 
     def emit(self, event: str, data: BaseSendMsg = {}, broadcast: bool = False, unicast_to: int = None):
         '''
@@ -213,9 +215,9 @@ class Connector:
         ----------
         event : str
             the event name 
-        
+
         data : BaseSendMsg
-            the data to send, fields 'timeStamp' and 'deviceId' are added when they are not present
+            the data to send, fields 'time_stamp' and 'device_id' are added when they are not present
 
         Optional
         --------
@@ -225,11 +227,11 @@ class Connector:
         unicast_to : int
             the device number to which this message is sent exclusively. When set, boradcast has no effect. 
         '''
-        if 'timeStamp' not in data:
-            data['timeStamp'] = time.time_ns() // 1000000
+        if 'time_stamp' not in data:
+            data['time_stamp'] = time.time_ns() // 1000000
 
-        if 'deviceId' not in data:
-            data['deviceId'] = self.device_id
+        if 'device_id' not in data:
+            data['device_id'] = self.device_id
 
         if broadcast:
             data['broadcast'] = True
@@ -237,8 +239,8 @@ class Connector:
         if type(unicast_to) == int:
             if 'broadcast' in data:
                 del data['broadcast']
-            data['unicastTo'] = unicast_to
-        
+            data['unicast_to'] = unicast_to
+
         self.sio.emit(event, data)
 
     def broadcast(self, data: DataMsg):
@@ -263,12 +265,16 @@ class Connector:
     def device_count(self) -> int:
         return len(self.devices)
 
+    @property
+    def client_devices(self) -> List[Device]:
+        return list(filter(lambda device: device['is_controller'], self.devices))
+
     @property
     def client_count(self) -> int:
         '''
         number of web-clients (or controller-clients) connected to this room
         '''
-        return len(list(filter(lambda device: device['isController'], self.devices)))
+        return len(self.client_devices)
 
     @property
     def joined_room_count(self) -> int:
@@ -285,7 +291,7 @@ class Connector:
         data = flatten(self.data.values())
         data = sorted(
             data,
-            key=lambda item: item['timeStamp'] if 'timeStamp' in item else 0,
+            key=lambda item: item['time_stamp'] if 'time_stamp' in item else 0,
             reverse=True
         )
         data = filter(lambda item: 'broadcast' in item and item['broadcast'], data)
@@ -305,7 +311,7 @@ class Connector:
         data = flatten(self.data.values())
         data = sorted(
             data,
-            key=lambda item: item['timeStamp'] if 'timeStamp' in item else 0,
+            key=lambda item: item['time_stamp'] if 'time_stamp' in item else 0,
             reverse=True
         )
 
@@ -399,7 +405,6 @@ class Connector:
         ])
         ```
         '''
-        did = self.device_id if (device_id is None) else device_id
         self.emit(
             ADD_NEW_DATA,
             {
@@ -447,16 +452,38 @@ class Connector:
             unicast_to=unicast_to
         )
 
+    def clean_data(self):
+        '''
+        removes all gathered data
+        '''
+        self.data = {}
+
+    def sleep(self, seconds=0) -> None:
+        '''
+        Sleep for the requested amount of time using the appropriate async model.
+
+        This is a utility function that applications can use to put a task to sleep without having to worry about using the correct call for the selected async mode.
+        '''
+        self.sio.sleep(seconds)
+
+    def wait(self):
+        '''
+        Wait until the connection with the server ends.
+
+        Client applications can use this function to block the main thread during the life of the connection.
+        '''
+        self.sio.wait()
+
     def disconnect(self):
         if not self.sio.connected:
             return
         self.sio.disconnect()
 
-    def join_room(self, deviceId: str):
-        self.emit(JOIN_ROOM, {'room': deviceId})
+    def join_room(self, device_id: str):
+        self.emit(JOIN_ROOM, {'room': device_id})
 
-    def leave_room(self, deviceId: str):
-        self.emit(LEAVE_ROOM, {'room': deviceId})
+    def leave_room(self, device_id: str):
+        self.emit(LEAVE_ROOM, {'room': device_id})
 
     def __on_connect(self):
         logging.info('SocketIO connected')
@@ -467,47 +494,60 @@ class Connector:
     def __register(self):
         self.emit(NEW_DEVICE)
 
+
+    def __callback(self, name, data):
+        callback = getattr(self, name)
+        if callback is None:
+            return
+        try:
+            arg_count = len(signature(callback).parameters)
+            if arg_count == 1:
+                callback(data)
+            elif arg_count == 2:
+                callback(data, self)
+        except Exception:
+            pass
+
     def __on_new_data(self, data: DataMsg):
-        if 'deviceId' not in data:
+        if 'device_id' not in data:
             return
 
-        if data['deviceId'] not in self.data:
-            self.data[data['deviceId']] = []
+        if data['device_id'] not in self.data:
+            self.data[data['device_id']] = []
 
-        self.data[data['deviceId']].append(data)
+        self.data[data['device_id']].append(data)
 
         if 'type' in data:
-            if data['type'] == 'key' and self.on_key is not None:
-                self.on_key(data)
-            if data['type'] in ['acceleration', 'gyro'] and self.on_sensor is not None:
-                self.on_sensor(data)
-            if data['type'] == 'acceleration' and self.on_acceleration is not None:
-                self.on_acceleration(data)
-            if data['type'] == 'gyro' and self.on_gyro is not None:
-                self.on_gyro(data)
-            if data['type'] == 'pointer' and self.on_pointer is not None:
-                self.on_pointer(data)
+            if data['type'] == 'key':
+                self.__callback('on_key', data)
+            if data['type'] in ['acceleration', 'gyro']:
+                self.__callback('on_sensor', data)
+            if data['type'] == 'acceleration':
+                self.__callback('on_acceleration', data)
+            if data['type'] == 'gyro':
+                self.__callback('on_gyro', data)
+            if data['type'] == 'pointer':
+                self.__callback('on_pointer', data)
 
         if 'broadcast' in data and data['broadcast'] and self.on_broadcast_data is not None:
-            self.on_broadcast_data(data)
-        elif self.on_data is not None:
-            self.on_data(data)
+            self.__callback('on_broadcast_data', data)
+        else:
+            self.__callback('on_data', data)
 
     def __on_all_data(self, data: List[DataMsg]):
-        if 'deviceId' not in data:
+        if 'device_id' not in data:
             return
 
-        self.data[data['deviceId']] = data['allData']
-        if self.on_all_data is not None:
-            self.on_all_data(data)
+        self.data[data['device_id']] = data['all_data']
+        self.__callback('on_all_data', data)
 
     def __on_room_left(self, device: DeviceLeftMsg):
         if device['room'] == self.device_id:
             if device['device'] in self.room_members:
                 self.room_members.remove(device['device'])
-                if self.on_room_left is not None:
-                    self.on_room_left(device['device'])
-        elif device['device']['deviceId'] == self.device_id:
+                self.__callback('on_room_left', device['device'])
+
+        elif device['device']['device_id'] == self.device_id:
             if device['device'] in self.joined_rooms:
                 self.joined_rooms.remove(device['device'])
 
@@ -515,39 +555,35 @@ class Connector:
         if device['room'] == self.device_id:
             if device['device'] not in self.room_members:
                 self.room_members.append(device['device'])
-                if self.on_room_joined is not None:
-                    self.on_room_joined(device['device'])
-        elif device['device']['deviceId'] == self.device_id:
+                self.__callback('on_room_joined', device['device'])
+        elif device['device']['device_id'] == self.device_id:
             if device['device'] not in self.joined_rooms:
                 self.joined_rooms.append(device['device'])
 
     def __on_error(self, err: ErrorMsg):
         logging.warn(f'Error on Event {err.type}: {err.msg}')
 
-        if self.on_error is not None:
-            self.on_error(err)
+        self.__callback('on_error', err)
 
     def __on_device(self, device: Device):
-        if 'deviceId' not in device or 'socketId' not in device:
+        if 'device_id' not in device or 'socket_id' not in device:
             return
-        if self.sio.sid == device['socketId']:
+        if self.sio.sid == device['socket_id']:
             self.device = device
             self.emit(GET_ALL_DATA)
             if device not in self.room_members:
                 self.room_members.append(device)
-            if self.on_device is not None:
-                self.on_device(device)
-
+            self.__callback('on_device', device)
+    
     def __on_devices(self, devices: List[Device]):
         had_client_device = self.client_device is not None
         self.devices = devices
         if self.on_client_device:
             has_client_device = self.client_device is not None
             if (had_client_device and not has_client_device) or (has_client_device and not had_client_device):
-                self.on_client_device(self.client_device)
+                self.__callback('on_client_device', self.client_device)
 
-        if self.on_devices is not None:
-            self.on_devices(devices)
+        self.__callback('on_devices', devices)
 
 
 if __name__ == '__main__':
@@ -561,7 +597,7 @@ if __name__ == '__main__':
         ['black', 'white', 'black'],
     ], broadcast=True)
 
-    connector.on_key = lambda data: logging.info(f'on_key: {data}')
+    connector.on_key = lambda data, c: logging.info(f'on_key: {data}, len: {len(c.all_data())}')
     connector.on_broadcast_data = lambda data: logging.info(f'on_broadcast_data: {data}')
     connector.on_data = lambda data: logging.info(f'on_data: {data}')
     connector.on_all_data = lambda data: logging.info(f'on_all_data: {data}')
@@ -574,7 +610,7 @@ if __name__ == '__main__':
     connector.on_room_left = lambda data: logging.info(f'on_room_left: {data}')
     connector.on_pointer = lambda data: logging.info(f'on_pointer: {data}')
     connector.on_client_device = lambda data: logging.info(f'on_client_device: {data}')
-    connector.on_error = lambda data: logging.info(f'on_error: {data}')    
+    connector.on_error = lambda data: logging.info(f'on_error: {data}')
 
     time.sleep(1)
     print(connector.joined_room_count)
@@ -585,7 +621,7 @@ if __name__ == '__main__':
     print('data: ', connector.all_data())
     print('data: ', connector.all_data(data_type='grid'))
     print('latest data: ', connector.latest_data())
-    print('timestamp', to_datetime(connector.latest_data()))
+    print('time_stamp', to_datetime(connector.latest_data()))
     print('latest data: ', connector.latest_data(data_type='key'))
     print('broadcast data: ', connector.all_broadcast_data())
     print('broadcast data: ', connector.all_broadcast_data(data_type='grid'))
@@ -596,5 +632,5 @@ if __name__ == '__main__':
     print('cnt clients', connector.client_count)
     print('cnt joined rooms', connector.joined_room_count)
 
-    connector.sio.wait()
+    connector.sleep(2)
     connector.disconnect()
